// ColorDungeon.cpp : Defines the entry point for the application.
//

#include "stdafx.h"
#include "ColorDungeon.h"
#include "Paint.h"
#include "Globals.h"
#include <gdiplus.h>
#include <assert.h>
#include <math.h>
#include <time.h>
#include <windowsx.h>
using namespace Gdiplus;

#define MAX_LOADSTRING 100
const int MAXX = 640;
const int MAXY = 480;
const wchar_t* TITLE = L"ColorDungeon";
const wchar_t* WINCLASS = L"ColorDungeonClass";
bool appdone = false;
const int TIMESTEP = 100;
clock_t laststep;

Globals g_globals;

// Global Variables:
HINSTANCE hInst;								// current instance
TCHAR szTitle[MAX_LOADSTRING];					// The title bar text
TCHAR szWindowClass[MAX_LOADSTRING];			// the main window class name


void OnAskResize(LPARAM lparam) // for forbidding resizing
{
	MINMAXINFO* bounds = (MINMAXINFO*)lparam;
	bounds->ptMinTrackSize.x = MAXX;
	bounds->ptMaxTrackSize.x = MAXX;
	bounds->ptMinTrackSize.y = MAXY;
	bounds->ptMaxTrackSize.y = MAXY;
}

void OnClick(HWND hwnd, LPARAM lparam)
{
	int cx = GET_X_LPARAM(lparam);
	int cy = GET_Y_LPARAM(lparam);
	if(g_globals.mouseOverSword && g_globals.swordBombCount > 0 && !g_globals.swordBomb.inProgress) {
		g_globals.swordBomb.Create(Point(g_globals.myPosX, g_globals.myPosY));
		g_globals.swordBombCount--;
	}
}

void OnMouseMove(HWND hwnd, LPARAM lparam)
{
	int cx = GET_X_LPARAM(lparam);
	int cy = GET_Y_LPARAM(lparam);
	// Check if mouse is over sword icon
	if(cx>=15 && cx<=15+40 && cy>10 && cy<=10+41)
	{
		g_globals.mouseOverSword = true;
		InvalidateRect(hwnd, NULL, false);
	}
	else if(g_globals.mouseOverSword)
	{
		g_globals.mouseOverSword = false;
		InvalidateRect(hwnd, NULL, false);
	}
}

float DistBtwn(Point a, Point b)
{
	float sqdist = ((a.X-b.X)*(a.X-b.X)) + ((a.Y-b.Y)*(a.Y-b.Y));
	return sqrt(sqdist);
}

bool AreClose(Point a, Point b, float thresh)
{
	float sqdist = ((a.X-b.X)*(a.X-b.X)) + ((a.Y-b.Y)*(a.Y-b.Y));
	return sqdist<(thresh*thresh);
}

bool AreClose(PointF a, Point b, float thresh)
{
	float sqdist = ((a.X-b.X)*(a.X-b.X)) + ((a.Y-b.Y)*(a.Y-b.Y));
	return sqdist<(thresh*thresh);
}

void Normalize(PointF& pt)
{
	float mag = sqrt((pt.X*pt.X) + (pt.Y*pt.Y));
	pt.X /= mag;
	pt.Y /= mag;
}

void UpdateGameState()
{
	if(g_globals.gameState == NORMAL)
	{
		clock_t thisstep = clock();
		if((thisstep - laststep) < TIMESTEP) return;
		laststep = thisstep;

		// Check for death
		if(g_globals.health == 0)
		{
			g_globals.gameState = GAME_OVER;
			return;
		}

		const int moveSpeed = g_globals.moveSpeed;
		// Move the character
		if(g_globals.amMovingLeft && !g_globals.l.IsObstacle(g_globals.myPosX-moveSpeed, g_globals.myPosY)) 
		{
			g_globals.myPosX -= moveSpeed; // left
		}
		else if(g_globals.amMovingRight && !g_globals.l.IsObstacle(g_globals.myPosX+moveSpeed, g_globals.myPosY)) 
		{
			g_globals.myPosX += moveSpeed; // right
		}	
		if(g_globals.amMovingUp && !g_globals.l.IsObstacle(g_globals.myPosX, g_globals.myPosY-moveSpeed)) 
		{
			g_globals.myPosY -= moveSpeed; // up
		}
		else if(g_globals.amMovingDown && !g_globals.l.IsObstacle(g_globals.myPosX, g_globals.myPosY+moveSpeed)) 
		{
			g_globals.myPosY += moveSpeed; // down
		}

		// Check for stepping on the staircase
		if(AreClose(Point((g_globals.l.staircase.X+1)*g_globals.tileSize, (g_globals.l.staircase.Y+1)*g_globals.tileSize), Point(g_globals.myPosX, g_globals.myPosY), g_globals.tileSize/2))
		{
			if(g_globals.floorNum < 4)
			{
				g_globals.floorNum++;
				InitGame();
				GenerateLevel();
			}
			else {
				g_globals.gameState = VICTORY1;
			}
			return;
		}

		// check swordbomb
		if(g_globals.swordBomb.inProgress)
		{
			g_globals.swordBomb.Update();
		}

		// Check for picking up an item
		if(g_globals.checkForItemPickup)
		{
			for(unsigned int i=0; i<g_globals.drops.size(); i++)
			{
				if(AreClose(g_globals.drops[i].pos, Point(g_globals.myPosX, g_globals.myPosY), g_globals.tileSize))
				{
					// We're picking up the item
					const int id = g_globals.drops[i].id;
					if(id == 0) // gem
					{
						if(g_globals.myColor==2) g_globals.myColor=4;
						else if(g_globals.myColor==4) g_globals.myColor=2;
						g_globals.drops.erase(g_globals.drops.begin()+i);
						break;
					}
					else if(id == 1 && g_globals.swordBombCount < 3)
					{
						g_globals.swordBombCount++;
						g_globals.drops.erase(g_globals.drops.begin()+i);
						break;
					}
				}
			}
			g_globals.checkForItemPickup = false;
		}

		for(unsigned int i=0; i<g_globals.drops.size(); i++)
		{
			g_globals.drops[i].frame = (g_globals.drops[i].frame+1)%8;
		}
		if(g_globals.amMovingLeft || g_globals.amMovingRight || g_globals.amMovingUp || g_globals.amMovingDown) g_globals.myFrameNum = (g_globals.myFrameNum+1)%4;
		
		// Update orb state
		for(unsigned int i=0; i<g_globals.orbs.size(); i++)
		{
			// Is the orb opposite color, and near our player?
			Orb& orb = g_globals.orbs[i];
			if(orb.dead) continue;

			if((orb.color == 2 && g_globals.myColor==4 || orb.color == 4 && g_globals.myColor==2) &&
				AreClose(orb.pos, Point(g_globals.myPosX, g_globals.myPosY), g_globals.tileSize/2))
			{
				// Suffer damage
				g_globals.health -= 0.05;
				if(g_globals.health < 0) g_globals.health = 0;
			}
			

			// Don't deal with orbs that lie offscreen
			if(abs(orb.pos.X-g_globals.cameraX) > g_globals.maxx) continue;
			if(abs(orb.pos.Y-g_globals.cameraY) > g_globals.maxy) continue;

			// Killed by swordbomb?
			if(g_globals.swordBomb.inProgress && AreClose(orb.pos, g_globals.swordBomb.center, g_globals.swordBomb.explosionRadius)) {
				orb.dead = true;
				continue;
			}

			if(orb.intelligent)
			{
				orb.dir.X = g_globals.myPosX - orb.pos.X;
				orb.dir.Y = g_globals.myPosY - orb.pos.Y;
				Normalize(orb.dir);
			}
			orb.majorCountdown--;
			if(orb.majorCountdown <= 0)
			{
				if(orb.color == 0) {
					// change from white to red
					orb.color = CoinFlip()?1:3;
					orb.majorCountdown = 10;
				}
				else if(orb.color == 1) { // change to red fully
					orb.color = 2;
					orb.majorCountdown = RandBtwn(50, 100);
				}
				else if(orb.color == 2) { // done being red. change to white
					orb.color = 0;
					orb.majorCountdown = RandBtwn(50, 100);
				}
				else if(orb.color == 3) { // on way to blue
					orb.color = 4;
					orb.majorCountdown = RandBtwn(50, 100);
				}
				else if(orb.color == 4) { // on way to blue
					orb.color = 1;
					orb.majorCountdown = RandBtwn(50, 100);
				}
			}
			// If orb lies at a boundary, adjust its speed so that it goes in the opposite direction.
			const int thresh = 10;
			if(orb.pos.X < thresh) { // hit the left boundary
				orb.dir.X = abs(orb.dir.X);
			}
			if(orb.pos.Y < thresh) { // hit the upper boundary
				orb.dir.Y = abs(orb.dir.Y);
			}
			if(orb.pos.X > g_globals.l.width*g_globals.tileSize) { // hit the right side boundary
				orb.dir.X = -abs(orb.dir.X);
			}
			if(orb.pos.Y > g_globals.l.height*g_globals.tileSize) { // hit the bottom boundary
				orb.dir.Y = -abs(orb.dir.Y);
			}

			orb.flicker = !orb.flicker;
			orb.pos.X += orb.dir.X*orb.speed;
			orb.pos.Y += orb.dir.Y*orb.speed;
		}
	}
	else if(g_globals.gameState == GAME_OVER)
	{
		// might no t have to do anything here
	}
	else if(g_globals.gameState == VICTORY1)
	{
		// might not have to do anything here
	}
	else if(g_globals.gameState == VICTORY2)
	{
		// might not have to do anything here
	}
	else if(g_globals.gameState == SPLASH_SCREEN)
	{
	}
	else if(g_globals.gameState == TUTORIAL)
	{
	}
}



void OnKeyDown(HWND hwnd, WPARAM wparam)
{
	int key = (int)wparam;
	if(g_globals.gameState == NORMAL)
	{
		
		const int moveSpeed = g_globals.moveSpeed;
		if(key == 0x1B);
		else if(key == 32);
		else if(key == 37) {
			if(!g_globals.l.IsObstacle(g_globals.myPosX-moveSpeed, g_globals.myPosY)) 
			{
				//g_globals.myPosX -= moveSpeed; // left
				g_globals.checkForItemPickup = true;
				g_globals.myDir = DIR_W;
				g_globals.amIdle = false;
				g_globals.amMovingRight = false;
				g_globals.amMovingLeft = true;
			}
		}
		else if(key == 38) {
			if(!g_globals.l.IsObstacle(g_globals.myPosX, g_globals.myPosY-moveSpeed)) 
			{
				//g_globals.myPosY -= moveSpeed; // up
				g_globals.checkForItemPickup = true;
				g_globals.myDir = DIR_N;
				g_globals.amIdle = false;
				g_globals.amMovingDown = false;
				g_globals.amMovingUp = true;

			}
		}
		else if(key == 39) {
			if(!g_globals.l.IsObstacle(g_globals.myPosX+moveSpeed, g_globals.myPosY))
			{
				//g_globals.myPosX += moveSpeed; // right
				g_globals.checkForItemPickup = true;
				g_globals.myDir = DIR_E;
				g_globals.amIdle = false;
				g_globals.amMovingLeft = false;
				g_globals.amMovingRight = true;
			}
		}
		else if(key == 40) {
			if(!g_globals.l.IsObstacle(g_globals.myPosX, g_globals.myPosY+moveSpeed))
			{
				//g_globals.myPosY += moveSpeed; // down
				g_globals.checkForItemPickup = true;
				g_globals.myDir = DIR_S;
				g_globals.amIdle = false;
				g_globals.amMovingUp = false;
				g_globals.amMovingDown = true;
			}
		}
		else if(key == 'b' || key=='B')
		{
			if(g_globals.swordBombCount > 0 && !g_globals.swordBomb.inProgress) {
				g_globals.swordBomb.Create(Point(g_globals.myPosX, g_globals.myPosY));
				g_globals.swordBombCount--;
			}
		}
		UpdateGameState();
		InvalidateRect(hwnd, NULL, false);
	}
	else if(g_globals.gameState == GAME_OVER)
	{
		InitGame();
		g_globals.floorNum = 1;
		g_globals.gameState = SPLASH_SCREEN;
	}
	else if(g_globals.gameState == VICTORY1)
	{
		g_globals.gameState = VICTORY2;
	}
	else if(g_globals.gameState == VICTORY2)
	{
		exit(0);
	}
	else if(g_globals.gameState == SPLASH_SCREEN)
	{
		if(key == 0x1B) g_globals.gameState = TUTORIAL;
		else g_globals.gameState = NORMAL;
	}
	else if(g_globals.gameState == TUTORIAL)
	{
		g_globals.gameState = SPLASH_SCREEN;
	}
}


void OnKeyUp(HWND hwnd, WPARAM wparam)
{
	int key = (int)wparam;
	g_globals.amIdle = true;
	if(key == 37) {
		g_globals.amMovingLeft = false;
		if(g_globals.amMovingUp) g_globals.myDir = DIR_N;
		else if(g_globals.amMovingDown) g_globals.myDir = DIR_S;

	}
	else if(key == 38) {
			g_globals.amMovingUp = false;
			if(g_globals.amMovingLeft) g_globals.myDir = DIR_W;
			else if(g_globals.amMovingRight) g_globals.myDir = DIR_E;
	}
	else if(key == 39) {
			g_globals.amMovingRight = false;
			if(g_globals.amMovingUp) g_globals.myDir = DIR_N;
			else if(g_globals.amMovingDown) g_globals.myDir = DIR_S;
	}
	else if(key == 40) {
			g_globals.amMovingDown = false;
			if(g_globals.amMovingLeft) g_globals.myDir = DIR_W;
			else if(g_globals.amMovingRight) g_globals.myDir = DIR_E;
	}
	if(!g_globals.amMovingLeft && !g_globals.amMovingRight && !g_globals.amMovingUp && !g_globals.amMovingDown) g_globals.myFrameNum = 0;
	UpdateGameState();
	InvalidateRect(hwnd, NULL, false);
}


void OnTimerTick(HWND hWnd)
{
	UpdateGameState();
	InvalidateRect(hWnd, NULL, false);
}

void OnPaint(HWND hwnd)
{
	HDC hdc;
	PAINTSTRUCT ps;
	Bitmap offscreen(MAXX, MAXY);
	Graphics graphics(&offscreen);
	graphics.SetSmoothingMode(SmoothingModeAntiAlias);
	Pen pen(Color(0, 0, 0));
	SolidBrush brush(Color(255, 255, 255));
	graphics.FillRectangle(&brush, 0, 0, MAXX, MAXY);
	PaintFrame(MAXX, MAXY, graphics);
	hdc = BeginPaint(hwnd, &ps);
	Graphics screen(hdc);
	CachedBitmap cb(&offscreen, &screen);
	screen.DrawCachedBitmap(&cb, 0, 0);
	EndPaint(hwnd, &ps);

}

// Forward declarations of functions included in this code module:
ATOM				MyRegisterClass(HINSTANCE hInstance);
BOOL				InitInstance(HINSTANCE, int);
LRESULT CALLBACK	WndProc(HWND, UINT, WPARAM, LPARAM);
INT_PTR CALLBACK	About(HWND, UINT, WPARAM, LPARAM);

int APIENTRY _tWinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPTSTR    lpCmdLine,
                     int       nCmdShow)
{
	
		g_globals.floorNum = 1;
		InitGame();
		GenerateLevel();

	g_globals.gameState = SPLASH_SCREEN;

	MSG msg;
	HWND hwnd;
	WNDCLASSEX wndclass;
	GdiplusStartupInput gdiplusStartupInput;
	ULONG_PTR gdiplusToken;
	GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, NULL);
	wndclass.cbSize = sizeof(WNDCLASSEX);
	wndclass.cbClsExtra = 0;
	wndclass.hInstance = hInstance;
	wndclass.cbWndExtra = 0;
	wndclass.style = CS_VREDRAW|CS_OWNDC;
	wndclass.lpfnWndProc = WndProc;
	wndclass.hIcon = LoadIcon(NULL, MAKEINTRESOURCE(IDI_APPLICATION));
	wndclass.hIconSm = LoadIcon(NULL, MAKEINTRESOURCE(IDI_APPLICATION));
	wndclass.hCursor = LoadCursor(NULL, IDC_ARROW);
	wndclass.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wndclass.lpszMenuName = NULL;
	wndclass.lpszClassName = WINCLASS;
	RegisterClassEx(&wndclass);
	hwnd = CreateWindow(WINCLASS, TITLE, WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, CW_USEDEFAULT, MAXX, MAXY, NULL, NULL, hInstance, NULL);
	assert(hwnd);
	
	// set timer
	SetTimer(hwnd, NULL, 100, NULL);

	ShowWindow(hwnd, SW_SHOWNORMAL);
	UpdateWindow(hwnd);
	while(GetMessage(&msg, hwnd, 0, 0)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
		if(appdone) break;
	}
	GdiplusShutdown(gdiplusToken);
	return (int)msg.wParam;
}



//
//  FUNCTION: MyRegisterClass()
//
//  PURPOSE: Registers the window class.
//
//  COMMENTS:
//
//    This function and its usage are only necessary if you want this code
//    to be compatible with Win32 systems prior to the 'RegisterClassEx'
//    function that was added to Windows 95. It is important to call this function
//    so that the application will get 'well formed' small icons associated
//    with it.
//
ATOM MyRegisterClass(HINSTANCE hInstance)
{
	WNDCLASSEX wcex;

	wcex.cbSize = sizeof(WNDCLASSEX);

	wcex.style			= CS_HREDRAW | CS_VREDRAW;
	wcex.lpfnWndProc	= WndProc;
	wcex.cbClsExtra		= 0;
	wcex.cbWndExtra		= 0;
	wcex.hInstance		= hInstance;
	wcex.hIcon			= LoadIcon(hInstance, MAKEINTRESOURCE(IDI_COLORDUNGEON));
	wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground	= (HBRUSH)(COLOR_WINDOW+1);
	wcex.lpszMenuName	= MAKEINTRESOURCE(IDC_COLORDUNGEON);
	wcex.lpszClassName	= szWindowClass;
	wcex.hIconSm		= LoadIcon(wcex.hInstance, MAKEINTRESOURCE(IDI_SMALL));

	return RegisterClassEx(&wcex);
}

//
//   FUNCTION: InitInstance(HINSTANCE, int)
//
//   PURPOSE: Saves instance handle and creates main window
//
//   COMMENTS:
//
//        In this function, we save the instance handle in a global variable and
//        create and display the main program window.
//
BOOL InitInstance(HINSTANCE hInstance, int nCmdShow)
{

   HWND hWnd;

   hInst = hInstance; // Store instance handle in our global variable

   hWnd = CreateWindow(szWindowClass, szTitle, WS_OVERLAPPEDWINDOW,
      CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, NULL, NULL, hInstance, NULL);

   if (!hWnd)
   {
      return FALSE;
   }

   ShowWindow(hWnd, nCmdShow);
   UpdateWindow(hWnd);

   return TRUE;
}

//
//  FUNCTION: WndProc(HWND, UINT, WPARAM, LPARAM)
//
//  PURPOSE:  Processes messages for the main window.
//
//  WM_COMMAND	- process the application menu
//  WM_PAINT	- Paint the main window
//  WM_DESTROY	- post a quit message and return
//
//
LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	int wmId, wmEvent;
	PAINTSTRUCT ps;
	HDC hdc;

	switch (message)
	{
	case WM_COMMAND:
		wmId    = LOWORD(wParam);
		wmEvent = HIWORD(wParam);
		// Parse the menu selections:
		switch (wmId)
		{
		case IDM_ABOUT:
			DialogBox(hInst, MAKEINTRESOURCE(IDD_ABOUTBOX), hWnd, About);
			break;
		case IDM_EXIT:
			DestroyWindow(hWnd);
			break;
		default:
			return DefWindowProc(hWnd, message, wParam, lParam);
		}
		break;
	case WM_PAINT: OnPaint(hWnd); break;
	case WM_MOUSEMOVE: OnMouseMove(hWnd, lParam); break;
	case WM_LBUTTONDOWN: OnClick(hWnd, lParam); break;
	case WM_DESTROY:
		PostQuitMessage(0);
		//FreePaint();
		appdone = true;
		break;
	case WM_KEYDOWN: OnKeyDown(hWnd, wParam); break;
	case WM_KEYUP: OnKeyUp(hWnd, wParam); break;	
	case WM_GETMINMAXINFO: OnAskResize(lParam); break;
	case WM_TIMER: OnTimerTick(hWnd); break;
	default:
		return DefWindowProc(hWnd, message, wParam, lParam);
	}
	return 0;
}

// Message handler for about box.
INT_PTR CALLBACK About(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	UNREFERENCED_PARAMETER(lParam);
	switch (message)
	{
	case WM_INITDIALOG:
		return (INT_PTR)TRUE;

	case WM_COMMAND:
		if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL)
		{
			EndDialog(hDlg, LOWORD(wParam));
			return (INT_PTR)TRUE;
		}
		break;
	}
	return (INT_PTR)FALSE;
}
