#pragma once
#include <vector>
#include <gdiplus.h>
using namespace Gdiplus;

enum Dir {DIR_N, DIR_E, DIR_S, DIR_W};

enum GameState {NORMAL, GAME_OVER, VICTORY1, VICTORY2, SPLASH_SCREEN, TUTORIAL};

struct Level
{
	void Set(int x, int y, unsigned int tile)
	{
		tiles[y*width + x] = tile;
	}

	unsigned int Get(int x, int y)
	{
		return tiles[y*width + x];
	}

	void MarkReachability(int x, int y, std::vector<bool>& matrix)
	{
	   matrix[y*width+x] = true;
	   if(x > 1 && Get(x-1, y) == 0 && !matrix[y*width + (x-1)]) MarkReachability(x-1, y, matrix);
	   if(x < width-2 && Get(x+1, y) == 0 && !matrix[y*width + (x+1)]) MarkReachability(x+1, y, matrix);
	   if(y > 1 && Get(x, y-1) == 0 && !matrix[(y-1)*width + (x)]) MarkReachability(x, y-1, matrix);
	   if(y < height-2 && Get(x, y+1) == 0 && !matrix[(y+1)*width + (x)]) MarkReachability(x, y+1, matrix);
	}

	// 0 --> empty space
	// 1 --> wall

	bool IsObstacle(int worldX, int worldY);

	Point staircase;
	std::vector<unsigned int> tiles;
	int width;
	int height;
};

float DistBtwn(Point a, Point b);

class Enemy
{
	Point pos;
	int width, height;

	void Update()
	{
		// Enemies move randomly but obey the walls of the dungeon.
	}
};

struct Drop
{
	Drop(): frame(0) {}
	Point pos;
	int width, height;
	int id; // what the item is
	int frame;
};

struct Orb
{
	int color; 
	// 0 --> white
	// 1 --> on way to red
	// 2 --> red
	// 3 --> on way to blue
	// 4 --> blue
	int majorCountdown; // countdown till next change
	bool flicker;
	PointF pos;
	PointF dir;
	float speed;
	bool intelligent;
	bool dead;
};

int RandBtwn(int lo, int hi);
bool CoinFlip(float p);
bool CoinFlip();

struct SwordBomb
{
	bool inProgress;
	Point center;
	struct Sword
	{
		Point pos;
		float height;
		float speed;
	};
	std::vector<Sword> swords;
	float explosionRadius;
	float explosionDir;
	void Create(Point center)
	{
		this->center = center;
		const float maxRadius = 200;
		explosionDir = true;
		// Place random swords because it looks cool
		for(unsigned int y = center.Y-maxRadius; y<center.Y+maxRadius; y+=RandBtwn(10, 20))
		{
			for(unsigned int x = center.X-maxRadius; x<center.X+maxRadius; x+=RandBtwn(10, 20))
			{
				if(DistBtwn(Point(x, y), center) > maxRadius) continue;
				if(!CoinFlip(0.05)) continue;
				Sword sword;
				sword.pos = Point(x, y);
				sword.height = (float)RandBtwn(200, 300);
				sword.speed = RandBtwn(20, 40);
				swords.push_back(sword);
			}
		}
		inProgress = true;
	}
	void Update()
	{
 		if(explosionDir && explosionRadius < 200) explosionRadius += 80;
		else { 
			explosionDir = false; 
			explosionRadius -= 10;
			if(explosionRadius < 0) explosionRadius = 0;
		}
		int inactive = 0;
		for(unsigned int i=0; i<swords.size(); i++)
		{
			swords[i].height -= swords[i].speed;
			if(swords[i].height <= 0) { swords[i].height = 0; inactive++;}
		}
		if(inactive == swords.size()) {
			inProgress = false;
			swords.clear();
		}
	}
	void Paint(Graphics& graphics);
};

struct Globals
{
	int myPosX, myPosY;
	Dir myDir;
	int myFrameNum;
	int myColor;
	int cameraX, cameraY;
	int tileSize;
	bool amIdle;
	float maxx, maxy;
	Level l;
	std::vector<Enemy> enemies;
	std::vector<Drop> drops;
	std::vector<Orb> orbs;
	bool checkForItemPickup;
	float health;
	int moveSpeed;
	GameState gameState;
	unsigned int swordBombCount;
	bool mouseOverSword;
	SwordBomb swordBomb;
	float difficulty; // a percentage
	int floorNum; // which floor we're on. affects difficulty
	

	bool amMovingLeft;
	bool amMovingRight;
	bool amMovingUp;
	bool amMovingDown;
};
extern Globals g_globals;

void GenerateLevel();
void InitGame();