#include "stdafx.h"
#include "Globals.h"
#include <gdiplus.h>
#include <vector>
#include <time.h>
#include <math.h>
using namespace Gdiplus;

const float PI = 3.14159;

int RandBtwn(int lo, int hi) // generates a random number between lo and hi, inclusive
{
	if (lo >= hi) return lo;
	int d = hi - lo + 1;
	return lo + rand()%d;
}

PointF RandomDir()
{
	int num = rand()%100;
	float percent = (float)num/100;
	float angle = 2 * PI * percent;
	PointF ret;
	ret.X = cos(angle);
	ret.Y = sin(angle);
	return ret;
}

bool CoinFlip(float p)
{
	if(p>=1) return true;
	int i = rand() % 1000;
	return i < 1000.0*p;
}

bool CoinFlip()
{
	return rand()%2==0;
}

bool Level::IsObstacle(int worldX, int worldY)
{
	const int tileX = worldX / g_globals.tileSize;
	const int tileY = worldY / g_globals.tileSize;
	if(tileX < 0 || tileX >= width ||
		tileY < 0 || tileY >= height ||
		Get(tileX, tileY) == 1) return true;
	return false;
}


void InitGame()
{
	srand(time(0));
	g_globals.cameraX = 0;
	g_globals.cameraY = 0;
	g_globals.maxx = 640;
	g_globals.maxy = 480;
	g_globals.tileSize = 64;
	g_globals.myPosX = g_globals.tileSize*2;
	g_globals.myPosY = g_globals.tileSize*2;
	g_globals.checkForItemPickup = true;
	g_globals.myDir = DIR_S;
	g_globals.myColor = 2;
	g_globals.myFrameNum = 0;
	g_globals.amIdle = true;
	g_globals.health = 1.0f;
	g_globals.moveSpeed = 15;
	g_globals.swordBombCount = 0;
	g_globals.mouseOverSword = false;
	
	if(g_globals.floorNum == 1) g_globals.difficulty = 0.1;
	else if(g_globals.floorNum == 2) g_globals.difficulty = 0.3;
	else if(g_globals.floorNum == 3) g_globals.difficulty = 0.5;
	else if(g_globals.floorNum == 4) g_globals.difficulty = 0.7;
	
	g_globals.amMovingLeft = false;
	g_globals.amMovingRight = false;
	g_globals.amMovingUp = false;
	g_globals.amMovingDown = false;
}

struct Room
{
	// left and right *indices*. *inclusive*.
	int left, right, top, bottom;
	bool split;
	std::vector<Point> exits;
	int width() {return right-left;}
	int height() {return bottom-top;}
};

void FormExits(Room& r)
{
	// top exit
	if(CoinFlip(0.3) && r.top>0)
	{
		int xPos = RandBtwn(r.left+1, r.right-1);
		r.exits.push_back(Point(xPos, r.top));
	}
	// btm exit
	if(CoinFlip(0.3) && r.bottom<g_globals.l.height-1)
	{
		int xPos = RandBtwn(r.left+1, r.right-1);
		r.exits.push_back(Point(xPos, r.bottom));
	}
	// left exit
	if(CoinFlip(0.3) && r.left>0)
	{
		int yPos = RandBtwn(r.top+1, r.bottom-1);
		r.exits.push_back(Point(r.left,yPos));
	}
	// right exit
	if(CoinFlip(0.3) && r.right<g_globals.l.width-1)
	{
		int yPos = RandBtwn(r.top+1, r.bottom-1);
		r.exits.push_back(Point(r.right,yPos));
	}
}

void Split(std::vector<Room>& rooms)
{
	const int maxWidth = 15;
	const int maxHeight = 15;
	for(unsigned int i=0; i<rooms.size(); i++)
	{
		if(rooms[i].width() > maxWidth && rooms[i].height()>maxHeight && CoinFlip(0.1) && !rooms[i].split)
		{
			// really need to partition the room
			rooms[i].split=true;
			Room newRoom;
			newRoom.split = false;
			newRoom.left = rooms[i].left + RandBtwn(4, maxWidth/2-1);
			newRoom.right = rooms[i].right - RandBtwn(4, maxWidth/2-1);
			newRoom.top = rooms[i].top + RandBtwn(4, maxWidth/2-1);
			newRoom.bottom = rooms[i].bottom - RandBtwn(4, maxWidth/2-1);
			FormExits(newRoom);
			// now we need to partition this inner room.
			std::vector<Room> newRoomList;
			newRoomList.push_back(newRoom);
			Split(newRoomList);
			for(unsigned int j=0; j<newRoomList.size(); j++) rooms.push_back(newRoomList[j]);
			// Sanity check
			if(newRoom.left <= rooms[i].left+1 || newRoom.right >=rooms[i].right-1 ||
				newRoom.top <= rooms[i].top+1 || newRoom.bottom >=rooms[i].bottom-1)
			{
				int j=5;
			}
		}
		else if(rooms[i].width() > maxWidth && !rooms[i].split)
		{
			// Make the room less wide by dividing it in half
			rooms[i].split = true;
			int SplitPt = RandBtwn(rooms[i].left+2, rooms[i].right-2);
			Room leftRoom = {rooms[i].left, SplitPt-1, rooms[i].top, rooms[i].bottom};
			leftRoom.split = false;
			FormExits(leftRoom);
			std::vector<Room> newRoomList1;
			newRoomList1.push_back(leftRoom);
			Split(newRoomList1);
			Room rightRoom = {SplitPt+1, rooms[i].right, rooms[i].top, rooms[i].bottom};
			rightRoom.split = false;
			FormExits(rightRoom);
			std::vector<Room> newRoomList2;
			newRoomList2.push_back(rightRoom);
			Split(newRoomList2);
			for(unsigned int j=0; j<newRoomList1.size(); j++) rooms.push_back(newRoomList1[j]);
			for(unsigned int j=0; j<newRoomList2.size(); j++) rooms.push_back(newRoomList2[j]);
		}
		else if(rooms[i].height() > maxHeight && !rooms[i].split)
		{
			// Make the room less high by dividing it in half
			rooms[i].split = true;
			int SplitPt = RandBtwn(rooms[i].top+2, rooms[i].bottom-2);
			Room topRoom = {rooms[i].left, rooms[i].right, rooms[i].top, SplitPt-1};
			FormExits(topRoom);
			topRoom.split = false;
			std::vector<Room> newRoomList1;
			newRoomList1.push_back(topRoom);
			Split(newRoomList1);
			Room bottomRoom = {rooms[i].left, rooms[i].right, SplitPt+1, rooms[i].bottom};
			FormExits(bottomRoom);
			bottomRoom.split = false;
			std::vector<Room> newRoomList2;
			newRoomList2.push_back(bottomRoom);
			Split(newRoomList2);
			for(unsigned int j=0; j<newRoomList1.size(); j++) rooms.push_back(newRoomList1[j]);
			for(unsigned int j=0; j<newRoomList2.size(); j++) rooms.push_back(newRoomList2[j]);
		}
	}
}

void GenerateLevel()
{
	Level& l = g_globals.l;
	l.width = RandBtwn(20, 40+(20.0*g_globals.difficulty));
	l.height = RandBtwn(20, 40+(20.0*g_globals.difficulty));

	// clear level to blank
	l.tiles.clear();
	for(unsigned int y=0; y<l.height; y++)
	{
		for(unsigned int x=0; x<l.width; x++)
		{
			l.tiles.push_back(0);
		}
	}
	std::vector<Room> rooms;
	Room room = {0, l.width-1, 0, l.height-1};
	room.split = false;
	rooms.push_back(room);
	Split(rooms);

	for(unsigned int i=0; i<rooms.size(); i++)
	{
		for(unsigned int j=rooms[i].left; j<=rooms[i].right; j++) 
		{
			l.Set(j, rooms[i].top, 1);
			l.Set(j, rooms[i].bottom, 1);
		}
		for(unsigned int j=rooms[i].top; j<=rooms[i].bottom; j++) 
		{
			l.Set(rooms[i].left, j, 1);
			l.Set(rooms[i].right, j, 1);
		}
		for(unsigned int j=0; j<rooms[i].exits.size(); j++)
		{
			l.Set(rooms[i].exits[j].X, rooms[i].exits[j].Y, 0);
		}
	}

	// Clear a corridor in the top left
	for(int i=1; i<4; i++)
	{
		for(int j=1; j<4; j++)
		{
			l.Set(i, j, 0);
		}
	}

	// Final reachability check

	// Now we have the reachability, what do we do about it?
	// Find the first unreachable tile, and clear a big X.
	bool allReached = false;
	while(!allReached)
	{
		std::vector<bool> reachability;
		reachability.resize(l.width*l.height, false);
		l.MarkReachability(1, 1, reachability);
		bool foundFirst=false;
		Point unreachable;
		for(unsigned int y=1; y<l.height-1; y++)
		{
			for(unsigned int x=1; x<l.width-1; x++)
			{
				if(l.Get(x, y)==0 && !reachability[y*l.width+x]) {
					unreachable = Point(x, y);
					foundFirst = true;
					break;
				}
			}
			if(foundFirst) break;
		}
		if(foundFirst) {
			for(unsigned int y=1; y<l.height-1; y++) l.Set(unreachable.X, y, 0);
			for(unsigned int x=1; x<l.width-1; x++) l.Set(x, unreachable.Y, 0);
		}
		else allReached = true;
	}

	// Insert items in free tiles
	for(unsigned int y=1; y<l.height-1; y++)
	{
		for(unsigned int x=1; x<l.width-1; x++)
		{
			if(l.Get(x, y) == 0)
			{
				if(CoinFlip(0.01 + (g_globals.difficulty * .03)))
				{
					// Add a gem
					Drop d;
					d.id = 0;
					d.pos.X = x * g_globals.tileSize + g_globals.tileSize/2;
					d.pos.Y = y * g_globals.tileSize + g_globals.tileSize/2;
					g_globals.drops.push_back(d);
				}
				else if(CoinFlip(0.01 + (g_globals.difficulty * .03)))
				{
					// Add a sword bomb
					Drop d;
					d.id = 1;
					d.pos.X = x * g_globals.tileSize + g_globals.tileSize/2;
					d.pos.Y = y * g_globals.tileSize + g_globals.tileSize/2;
					g_globals.drops.push_back(d);
				}
				// Add an orb
				else if( x > 4 && y > 4 && CoinFlip(0.2 + g_globals.difficulty*0.2))
				{
					Orb o;
					o.pos.X = x * g_globals.tileSize + g_globals.tileSize/2;
					o.pos.Y = y * g_globals.tileSize + g_globals.tileSize/2;
					o.dir = RandomDir();
					o.speed = (float)RandBtwn(1, 3) + g_globals.difficulty*RandBtwn(1, 3);;
					o.intelligent = CoinFlip(g_globals.difficulty+0.2);
					o.color = 0;
					if(CoinFlip(0.2)) // make this a red orb
					{
						o.color = 2;
						o.majorCountdown = RandBtwn(20 - (g_globals.difficulty*10), 40 - (g_globals.difficulty*20));
					}
					else if(CoinFlip(0.2))
					{
						o.color = 4;
						o.majorCountdown = RandBtwn(20 - (g_globals.difficulty*10), 40 - (g_globals.difficulty*20));
					}
					o.flicker = false;
					o.dead = false;
					g_globals.orbs.push_back(o);
				}
			}
		}
	}

	// Place the staircase somewhere near the lower right
	bool placed = false;
	for(int y=l.height-2; y>=0; y--)
	{
		for(int x=l.width-2; x>=0; x--)
		{
			if(l.Get(x, y) == 0) 
			{
				// We can place the staircase here.
				l.Set(x, y, 2);
				l.staircase = Point(x, y);
				placed=true;
				break;
			}
		}
		if(placed) break;
	}

}